/**
 * Program      esp32NtpTime.cpp
 * Author       2022-05-04 Charles Geiser (https://www.dodeka.ch)
 * 
 * Purpose      The program shows how to get the time from an NTP Server Pool.
 *              Thanks to the function configTzTime() we don't have to worry 
 *              about the connection to an NTP server and the time zone 
 *              difference to UTC. We always get the correct local time, even 
 *              the change from normal to daylight saving time is automatic.
 *              The internal RTC (real time clock) of the ESP32 is automatically
 *              synchronized. 
 *              After calling configTzTime() we could close the WiFi connection. 
 *              But because the internal reference frequency of the ESP32 can be 
 *              inaccurate and temperature dependent, we keep the WiFi connection 
 *              and leave the periodic synchronisation also to the internals of
 *              configTzTime().
 * 
 * Board        ESP32 DoIt DevKit V1
 *
 * Wiring 
 *
 * Remarks
 *
 * References   https://www.gnu.org/software/libc/manual/html_node/TZ-Variable.html
 */
#include <Arduino.h>
#include <WiFi.h>

const char HOSTNAME[]        ="esp-time";
const char NTP_SERVER_POOL[] = "ch.pool.ntp.org";
const char TIME_ZONE[]       = "MEZ-1MESZ-2,M3.5.0/02:00:00,M10.5.0/03:00:00";

// Define an enum for the various formats
enum class TF { TIME_ONLY, 
                DATE_ONLY, 
                DATE_TIME, 
                DATE_TIME_FULL, 
                DATE_TIME_ABBR, 
                DATE_TIME_ZONE,
                DATE_TIME_US,
                NOP };

// WiFi credentials 
const char ssid[]     = "Dodeka2G4";
const char password[] = "5408EnnetbadenHoehtalstrasse13";

const uint32_t secondsToSynchronize = 5;

/**
 * Print date and time in various formats
 */
void printDateTime(TF format)
{
  tm   rtcTime;
  char buf[40];
  int  bufSize = sizeof(buf);

  getLocalTime(&rtcTime);
  switch (format)
  {
    case TF::TIME_ONLY:  
      strftime(buf, bufSize, "%T",                &rtcTime); // hh:mm:ss
    break;
    case TF::DATE_ONLY:  
      strftime(buf, bufSize, "%F",                &rtcTime); // YYYY-MM-DD
    break;
    case TF::DATE_TIME:  
      strftime(buf, bufSize, "%F %T",             &rtcTime); // 2019-01-15 16:33:20
    break;
    case TF::DATE_TIME_FULL:  
      strftime(buf, bufSize, "%B %d %Y %T (%A)",  &rtcTime); // January 15 2019 16:33:20 (Tuesday)
    break;
    case TF::DATE_TIME_ABBR:  
      strftime(buf, bufSize, "%c",                &rtcTime); // Tue Jan 15 16:33:20 2019
    break;
    case TF::DATE_TIME_ZONE:  
      strftime(buf, bufSize, "%F %T %W/%w %Z %z", &rtcTime); // 2019-01-15 16:51:18 02/2 MEZ +0100
    break;
    case TF::DATE_TIME_US: 
      strftime(buf, bufSize, "%D %r",             &rtcTime ); // 08/23/01 02:55:02 pm
    break;
    default:
    break;
  }
  Serial.printf("%s\n", buf);
}

/**
 * Use a raw string literal to print a formatted
 * string of WiFi connection details
 */
void printConnectionDetails()
{
  Serial.printf(R"(
Connection Details:
------------------
  SSID       : %s
  Hostname   : %s
  IP-Address : %s
  MAC-Address: %s
  RSSI       : %d (received signal strength indicator)

)", WiFi.SSID().c_str(),
      //WiFi.hostname().c_str(),  // ESP8266
      WiFi.getHostname(),    // ESP32 
      WiFi.localIP().toString().c_str(),
      WiFi.macAddress().c_str(),
      WiFi.RSSI());  
}

/**
 * Establish the WiFi connection
 */
void initWiFi()
{
  Serial.println("Connecting to WiFi");
  WiFi.setHostname(HOSTNAME);
  WiFi.begin(ssid, password);

  // Try forever
  while (WiFi.status() != WL_CONNECTED) 
  {
    Serial.println("...Connecting to WiFi");
    delay(1000);
  }
  Serial.println("Connected");
  printConnectionDetails();
  log_i("===> done");
}

void initNTP()
{
  configTzTime(TIME_ZONE, NTP_SERVER_POOL);
  log_i("===> done");
}

/**
 * Returns true, as soon as msWait milliseconds have passed.
 * Supply a reference to an unsigned long variable to hold 
 * the previous milliseconds.
 */
bool waitIsOver(uint32_t &msPrevious, uint32_t msWait)
{
  return (millis() - msPrevious > msWait) ? (msPrevious = millis(), true) : false;
}

/**
 * Synchronize the output to full seconds
 */
void synchronize(uint32_t toSeconds)
{
  tm rtcTime;
  bool synchronized = false;

  while (! synchronized) 
  {
    getLocalTime(&rtcTime);
    if (rtcTime.tm_sec % toSeconds == 0) 
    {  
      synchronized = true;
      printDateTime(TF::TIME_ONLY);
    }
  } 
}

void setup() 
{
  Serial.begin(115200);
  initWiFi();
  initNTP();
  synchronize(secondsToSynchronize);
}

void loop() 
{
  static uint32_t msPrevious = millis();

  if (waitIsOver(msPrevious, secondsToSynchronize * 1000)) 
  {
    for (int i = (int)TF::TIME_ONLY; i < (int)TF::NOP; i++)
      printDateTime((TF)i);

    Serial.println();
  } 
}